@@ -1,14 +1,6 @@
 import ctypes
 
-from ctypes import (
-    c_int,
-    c_float,
-    c_char_p,
-    c_void_p,
-    c_bool,
-    POINTER,
-    Structure,
-)
+from ctypes import c_int, c_float, c_char_p, c_void_p, c_bool, POINTER, Structure, Array
 
 import pathlib
 from itertools import chain
@@ -47,7 +39,6 @@ class llama_context_params(Structure):
         ("n_ctx", c_int),  # text context
         ("n_parts", c_int),  # -1 for default
         ("seed", c_int),  # RNG seed, 0 for random
-
         ("f16_kv", c_bool),  # use fp16 for KV cache
         (
             "logits_all",
@@ -56,7 +47,6 @@ class llama_context_params(Structure):
         ("vocab_only", c_bool),  # only load the vocabulary, no weights
         ("use_mlock", c_bool),  # force system to keep model in RAM
         ("embedding", c_bool),  # embedding mode only
-
         # called with a progress value between 0 and 1, pass NULL to disable
         ("progress_callback", llama_progress_callback),
         # context pointer passed to the progress callback
@@ -118,7 +108,7 @@ _lib.llama_model_quantize.restype = c_int
 # Returns 0 on success
 def llama_eval(
     ctx: llama_context_p,
-    tokens: llama_token_p,
+    tokens,  # type: Array[llama_token]
     n_tokens: c_int,
     n_past: c_int,
     n_threads: c_int,
@@ -138,7 +128,7 @@ _lib.llama_eval.restype = c_int
 def llama_tokenize(
     ctx: llama_context_p,
     text: bytes,
-    tokens: llama_token_p,
+    tokens,  # type: Array[llama_token]
     n_max_tokens: c_int,
     add_bos: c_bool,
 ) -> c_int:
@@ -197,7 +187,7 @@ _lib.llama_get_embeddings.restype = POINTER(c_float)
 
 
 # Token Id -> String. Uses the vocabulary in the provided context
-def llama_token_to_str(ctx: llama_context_p, token: int) -> bytes:
+def llama_token_to_str(ctx: llama_context_p, token: llama_token) -> bytes:
     return _lib.llama_token_to_str(ctx, token)
 
 
@@ -226,7 +216,7 @@ _lib.llama_token_eos.restype = llama_token
 # TODO: improve the last_n_tokens interface ?
 def llama_sample_top_p_top_k(
     ctx: llama_context_p,
-    last_n_tokens_data: llama_token_p,
+    last_n_tokens_data,  # type: Array[llama_token]
     last_n_tokens_size: c_int,
     top_k: c_int,
     top_p: c_float,
